import time
from typing import Any, Iterator

import structlog
from strawberry.extensions import SchemaExtension

from orchestrator.db import db

logger = structlog.get_logger()


class StatsExtension(SchemaExtension):
    """Gathers various statistics for an executed GraphQL query, which are logged and returned in the extension results.

    Currently gathered statistics:
    - Operation time of the strawberry query
    - Time spent waiting on SQLAlchemy queries
    - Number of executed SQLAlchemy queries
    """

    before: dict[str, Any] | None
    after: dict[str, Any] | None
    start: float | None
    end: float | None

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)
        self.before = None
        self.after = None
        self.start = None
        self.end = None

    def get_results(self) -> dict[str, Any]:
        if self.after is None or self.before is None:
            return {}

        estimated_queries = self.after["queries_completed"] - self.before.get("queries_completed", 0)
        estimated_query_time = self.after["query_time_spent"] - self.before.get("query_time_spent", 0.0)
        operation_time = self.end - self.start if (self.start is not None and self.end is not None) else "n/a"

        return {
            "stats": {
                "db_queries": estimated_queries,
                "db_time": estimated_query_time,
                "operation_time": operation_time,
            }
        }

    def on_operation(self, *args, **kwargs) -> Iterator[None]:  # type: ignore
        self.before = db.session.connection().info.copy()
        self.start = time.time()

        yield

        self.end = time.time()
        self.after = db.session.connection().info.copy()

        stats = self.get_results().get("stats", {})
        logger.info(
            "GraphQL query stats",
            query=self.execution_context.query,
            variables=self.execution_context.variables,
            **stats,
        )



# Copyright 2022-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Callable, Iterable
from http import HTTPStatus
from pathlib import Path
from typing import Any, Coroutine, Protocol

import strawberry
import structlog
from fastapi.routing import APIRouter
from graphql import GraphQLError
from httpx import HTTPStatusError
from strawberry.extensions import SchemaExtension
from strawberry.fastapi import GraphQLRouter
from strawberry.tools import merge_types
from strawberry.types import ExecutionContext
from strawberry.utils.logging import StrawberryLogger

from nwastdlib.graphql.extensions.deprecation_checker_extension import make_deprecation_checker_extension
from nwastdlib.graphql.extensions.error_handler_extension import ErrorHandlerExtension, ErrorType
from oauth2_lib.fastapi import AuthManager
from oauth2_lib.strawberry import authenticated_field
from orchestrator.domain.base import SubscriptionModel
from orchestrator.graphql.autoregistration import create_subscription_strawberry_type, register_domain_models
from orchestrator.graphql.extensions.model_cache import ModelCacheExtension
from orchestrator.graphql.extensions.stats import StatsExtension
from orchestrator.graphql.mutations.customer_description import CustomerSubscriptionDescriptionMutation
from orchestrator.graphql.mutations.start_process import ProcessMutation
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.resolvers import (
    SettingsMutation,
    resolve_customer,
    resolve_process,
    resolve_processes,
    resolve_product_blocks,
    resolve_products,
    resolve_resource_types,
    resolve_settings,
    resolve_subscription,
    resolve_subscriptions,
    resolve_version,
    resolve_workflows,
)
from orchestrator.graphql.resolvers.scheduled_tasks import resolve_scheduled_tasks
from orchestrator.graphql.schemas import DEFAULT_GRAPHQL_MODELS
from orchestrator.graphql.schemas.customer import CustomerType
from orchestrator.graphql.schemas.process import ProcessType
from orchestrator.graphql.schemas.product import ProductType
from orchestrator.graphql.schemas.product_block import ProductBlock
from orchestrator.graphql.schemas.resource_type import ResourceType
from orchestrator.graphql.schemas.scheduled_task import ScheduledTaskGraphql
from orchestrator.graphql.schemas.settings import StatusType
from orchestrator.graphql.schemas.subscription import SubscriptionInterface
from orchestrator.graphql.schemas.version import VersionType
from orchestrator.graphql.schemas.workflow import Workflow
from orchestrator.graphql.types import SCALAR_OVERRIDES, OrchestratorContext, ScalarOverrideType, StrawberryModelType
from orchestrator.services.process_broadcast_thread import ProcessDataBroadcastThread
from orchestrator.settings import app_settings

api_router = APIRouter()

logger = structlog.get_logger(__name__)


@strawberry.federation.type(description="Orchestrator queries")
class OrchestratorQuery:
    process: ProcessType | None = authenticated_field(resolver=resolve_process, description="Returns a single process")
    processes: Connection[ProcessType] = authenticated_field(
        resolver=resolve_processes, description="Returns list of processes"
    )
    products: Connection[ProductType] = authenticated_field(
        resolver=resolve_products, description="Returns list of products"
    )
    product_blocks: Connection[ProductBlock] = authenticated_field(
        resolver=resolve_product_blocks, description="Returns list of product blocks"
    )
    resource_types: Connection[ResourceType] = authenticated_field(
        resolver=resolve_resource_types, description="Returns list of resource types"
    )
    workflows: Connection[Workflow] = authenticated_field(
        resolver=resolve_workflows, description="Returns list of workflows"
    )
    subscription: SubscriptionInterface | None = authenticated_field(
        resolver=resolve_subscription, description="Returns a single Subscription"
    )
    subscriptions: Connection[SubscriptionInterface] = authenticated_field(
        resolver=resolve_subscriptions, description="Returns list of subscriptions"
    )
    settings: StatusType = authenticated_field(
        resolver=resolve_settings,
        description="Returns information about cache, workers, and global engine settings",
    )
    version: VersionType = authenticated_field(resolver=resolve_version, description="Returns version information")
    scheduled_tasks: Connection[ScheduledTaskGraphql] = authenticated_field(
        resolver=resolve_scheduled_tasks, description="Returns scheduled job information"
    )


@strawberry.federation.type(description="Orchestrator customer Query")
class CustomerQuery:
    customers: Connection[CustomerType] = authenticated_field(
        resolver=resolve_customer, description="Returns default customer information"
    )


Query = merge_types("Query", (OrchestratorQuery, CustomerQuery))
Mutation = merge_types("Mutation", (SettingsMutation, CustomerSubscriptionDescriptionMutation, ProcessMutation))

OrchestratorGraphqlRouter = GraphQLRouter


class OrchestratorSchema(strawberry.federation.Schema):
    def process_errors(
        self,
        errors: list[GraphQLError],
        execution_context: ExecutionContext | None = None,
    ) -> None:
        """Override error processing to reduce verbosity of the logging.

        https://strawberry.rocks/docs/types/schema#handling-execution-errors
        """
        for error in errors:
            error_type = error.extensions.get("error_type") if error.extensions else None
            if (
                isinstance(error.original_error, HTTPStatusError)
                and error.original_error.response.status_code == HTTPStatus.NOT_FOUND
            ):
                message = str(error.original_error).splitlines()[0]  # Strip "For more info"
                StrawberryLogger.logger.debug(message)
            elif error_type in (ErrorType.NOT_AUTHORIZED, ErrorType.NOT_AUTHENTICATED):
                StrawberryLogger.logger.info(error.message)
            else:
                StrawberryLogger.error(error, execution_context)


class ContextGetterFactory(Protocol):
    def __call__(
        self,
        auth_manager: AuthManager,
        graphql_models: StrawberryModelType,
        broadcast_thread: ProcessDataBroadcastThread | None = None,
    ) -> Callable[[], Coroutine[Any, Any, OrchestratorContext]]: ...


def default_context_getter(
    auth_manager: AuthManager,
    graphql_models: StrawberryModelType,
    broadcast_thread: ProcessDataBroadcastThread | None = None,
) -> Callable[[], Coroutine[Any, Any, OrchestratorContext]]:
    async def context_getter() -> OrchestratorContext:
        return OrchestratorContext(
            auth_manager=auth_manager, graphql_models=graphql_models, broadcast_thread=broadcast_thread
        )

    return context_getter


def get_extensions(mutation: Any, query: Any) -> Iterable[type[SchemaExtension]]:
    yield ModelCacheExtension
    yield ErrorHandlerExtension
    if app_settings.ENABLE_GRAPHQL_DEPRECATION_CHECKER:
        yield make_deprecation_checker_extension(query=query, mutation=mutation)
    if app_settings.ENABLE_GRAPHQL_STATS_EXTENSION:
        yield StatsExtension
    if app_settings.ENABLE_GRAPHQL_PROFILING_EXTENSION:
        from strawberry.extensions import pyinstrument

        yield pyinstrument.PyInstrument(report_path=Path("pyinstrument.html"))  # type: ignore


def create_graphql_router(
    auth_manager: AuthManager,
    query: Any = Query,
    mutation: Any = Mutation,
    register_models: bool = True,
    subscription_interface: type = SubscriptionInterface,
    broadcast_thread: ProcessDataBroadcastThread | None = None,
    graphql_models: StrawberryModelType | None = None,
    scalar_overrides: ScalarOverrideType | None = None,
    extensions: list | None = None,
    custom_context_getter: ContextGetterFactory | None = None,
) -> OrchestratorGraphqlRouter:
    scalar_overrides = scalar_overrides if scalar_overrides else dict(SCALAR_OVERRIDES)
    models = graphql_models if graphql_models else dict(DEFAULT_GRAPHQL_MODELS)
    models["subscription"] = create_subscription_strawberry_type(
        "Subscription", SubscriptionModel, subscription_interface
    )

    if register_models:
        models = register_domain_models(subscription_interface, existing_models=models)

    extensions = extensions or list(get_extensions(mutation, query))

    schema = OrchestratorSchema(
        query=query,
        mutation=mutation,
        enable_federation_2=app_settings.FEDERATION_ENABLED,
        types=tuple(models.values()),
        extensions=extensions,
        scalar_overrides=scalar_overrides,
    )

    context_getter_factory = custom_context_getter or default_context_getter
    return OrchestratorGraphqlRouter(
        schema,
        context_getter=context_getter_factory(auth_manager, models, broadcast_thread),  # type: ignore
        graphiql=app_settings.SERVE_GRAPHQL_UI,
    )



import structlog
from sqlalchemy import BinaryExpression

from orchestrator.db import ProductBlockTable, ProductTable
from orchestrator.db.filters import create_memoized_field_list, generic_filter_from_clauses
from orchestrator.db.filters.search_filters import default_inferred_column_clauses, filter_exact, node_to_str_val
from orchestrator.utils.search_query import Node

logger = structlog.get_logger(__name__)


def product_block_clause(node: Node) -> BinaryExpression:
    return ProductTable.product_blocks.any(ProductBlockTable.name.ilike(node_to_str_val(node)))


PRODUCT_TABLE_COLUMN_CLAUSES = default_inferred_column_clauses(ProductTable) | {
    "product_block": product_block_clause,
    "tag": filter_exact(ProductTable.tag),
}

product_filter_fields = create_memoized_field_list(PRODUCT_TABLE_COLUMN_CLAUSES)
filter_products = generic_filter_from_clauses(PRODUCT_TABLE_COLUMN_CLAUSES)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from fastapi import WebSocket, status
from starlette.concurrency import run_until_first_complete
from starlette.websockets import WebSocketDisconnect
from structlog.stdlib import BoundLogger, get_logger

from orchestrator.utils.json import json_dumps
from orchestrator.utils.redis import RedisBroadcast

logger = get_logger(__name__)


def _make_logger(websocket: WebSocket, channel: str) -> BoundLogger:
    return logger.bind(
        client=websocket.client, channel=channel, sec_websocket_key=websocket.headers.get("sec-websocket-key")
    )


class BroadcastWebsocketManager:
    def __init__(self, broadcast_url: str):
        self.connected: list[WebSocket] = []
        self.broadcast_url = broadcast_url
        self.broadcast = RedisBroadcast(broadcast_url)

    async def connect_redis(self) -> None:
        await self.broadcast.connect()

    async def disconnect_redis(self) -> None:
        await self.broadcast.disconnect()

    async def connect(self, websocket: WebSocket, channel: str) -> None:
        """Connect a new websocket client."""
        self.connected.append(websocket)
        log = _make_logger(websocket, channel)
        log.debug("Websocket client connected, start loop", total_connections=len(self.connected))
        try:
            await run_until_first_complete(
                (self.sender, {"websocket": websocket, "channel": channel}),
                (self.receiver, {"websocket": websocket, "channel": channel}),
            )
        except Exception as exc:  # noqa: S110
            log.info("Websocket client loop stopped with an exception", message=str(exc))
        else:
            log.debug("Websocket client loop stopped normally")
        self.remove_ws_from_connected_list(websocket)

    async def disconnect(
        self, websocket: WebSocket, code: int = status.WS_1000_NORMAL_CLOSURE, reason: dict | str | None = None
    ) -> None:
        if reason:
            await websocket.send_text(json_dumps(reason))
        await websocket.close(code)
        self.remove_ws_from_connected_list(websocket)

    async def disconnect_all(self) -> None:
        for websocket in self.connected:
            await self.disconnect(websocket, code=status.WS_1001_GOING_AWAY, reason="Shutting down")

    async def receiver(self, websocket: WebSocket, channel: str) -> None:
        """Read messages from websocket client."""
        log = _make_logger(websocket, channel)
        while True:
            try:
                message = await websocket.receive_text()
                log.debug("Received websocket message", message=repr(message))
                if message == "__ping__":
                    await websocket.send_text("__pong__")
            except WebSocketDisconnect as disconnect:
                log.debug("Websocket connection closed by client", code=disconnect.code, reason=disconnect.reason)
                break
            except Exception:
                log.exception("Exception in receiver loop")
                break

    async def sender(self, websocket: WebSocket, channel: str) -> None:
        """Read messages from redis channel and send to websocket client."""
        log = _make_logger(websocket, channel)

        def parse_message(raw_message: Any) -> str | None:
            match raw_message:
                case {"type": "message", "data": bytes() as data}:
                    return data.decode()
                case None:
                    return None
                case _:
                    log.info("Drop unrecognized message", raw=raw_message)
                    return None

        try:
            async with self.broadcast.subscriber(channel) as subscriber:
                log.debug("Websocket client subscribed to channel")
                while True:
                    raw = await subscriber.get_message(timeout=1)
                    if (message := parse_message(raw)) is None:
                        continue

                    log.debug("Send websocket message", message=message)
                    await websocket.send_text(message)
        except Exception:
            log.exception("Exception in sender loop")

    async def broadcast_data(self, channels: list[str], data: dict) -> None:
        """Send messages to redis channel.

        This can be called by API and/or Worker instances.
        """
        message = json_dumps(data)
        async with RedisBroadcast(self.broadcast_url).pipeline() as pipe:
            for channel in channels:
                pipe.publish(channel, message)

    def remove_ws_from_connected_list(self, websocket: WebSocket) -> None:
        if websocket in self.connected:
            self.connected.remove(websocket)
        logger.debug("Websocket client disconnected", total_connections=len(self.connected), client=websocket.client)



# Copyright 2019-2025 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the orchestrator workflow engine."""

__version__ = "4.4.0rc1"

from orchestrator.app import OrchestratorCore
from orchestrator.settings import app_settings
from orchestrator.workflow import begin, conditional, done, focussteps, inputstep, retrystep, step, steplens, workflow

__all__ = [
    "OrchestratorCore",
    "app_settings",
    "step",
    "inputstep",
    "workflow",
    "retrystep",
    "begin",
    "done",
    "conditional",
    "focussteps",
    "steplens",
]



from typing import Iterable

from graphql.pyutils import camel_to_snake
from more_itertools import unique_everseen
from strawberry.types.nodes import FragmentSpread, InlineFragment, SelectedField

from orchestrator.graphql.types import OrchestratorInfo


def _get_fields(elem: InlineFragment | FragmentSpread | SelectedField, *field_names: str) -> Iterable[str]:
    if not elem.selections:
        yield ".".join(field_names)
        return

    for selection in elem.selections:
        match selection:
            case InlineFragment() | FragmentSpread():
                yield from _get_fields(selection, *field_names)
            case SelectedField(name=field_name):
                yield from _get_fields(selection, *field_names, camel_to_snake(field_name))


def get_selected_paths(info: OrchestratorInfo) -> list[str]:
    """Get all selected paths in the GraphQL query.

    Example:
        ['page.name',
         'page.target',
         'page.products.tag',
         'page.products.product_blocks.name',
         'page.products.product_blocks.resource_types.description',
         'page.products.product_blocks.in_use_by.name',
         'page.created_at'
        ]
    """

    def get_all_paths() -> Iterable[str]:
        for selected_field in info.selected_fields:
            yield from _get_fields(selected_field)

    return list(unique_everseen(get_all_paths()))



"""Add postgres function subscription_instance_as_json.

Revision ID: 42b3d076a85b
Revises: bac6be6f2b4f
Create Date: 2025-03-06 15:03:09.477225

"""

from pathlib import Path

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "42b3d076a85b"
down_revision = "bac6be6f2b4f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()

    revision_file_path = Path(__file__)
    with open(revision_file_path.with_suffix(".sql")) as f:
        conn.execute(text(f.read()))


def downgrade() -> None:
    conn = op.get_bind()

    conn.execute(text("DROP FUNCTION IF EXISTS subscription_instance_as_json;"))
    conn.execute(text("DROP FUNCTION IF EXISTS subscription_instance_fields_as_json;"))



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements websocket endpoints."""

from typing import Annotated

import structlog
from fastapi import Header, status
from fastapi.routing import APIRouter
from fastapi.websockets import WebSocket

from orchestrator.settings import app_settings
from orchestrator.websocket import WS_CHANNELS, websocket_manager

logger = structlog.get_logger(__name__)

router = APIRouter()


WEBSOCKET_AUTH_SUBPROTOCOL = "base64.bearer.token"


class WebsocketAuthError(ValueError):
    pass


async def get_subprotocol_and_token(sec_websocket_protocol: str | None) -> tuple[str | None, str]:
    if not sec_websocket_protocol:
        return None, ""

    try:
        subprotocol, token = sec_websocket_protocol.split(", ")
    except ValueError:
        raise WebsocketAuthError("Sec-Websocket-Protocol should contain subprotocol and token separated by a comma")

    if subprotocol != WEBSOCKET_AUTH_SUBPROTOCOL:
        raise WebsocketAuthError(f"Sec-Websocket-Protocol subprotocol should be {WEBSOCKET_AUTH_SUBPROTOCOL}")

    return subprotocol, token


if app_settings.ENABLE_WEBSOCKETS:

    @router.websocket("/events")
    async def websocket_events(
        websocket: WebSocket, sec_websocket_protocol: Annotated[str | None, Header()] = None
    ) -> None:
        """Emits events for the frontend.

        Events are of the form {"name": <name>, "value": <value>}.
        To authenticate, provide an array [<WEBSOCKET_AUTH_SUBPROTOCOL>, <JWT token>] in the "Sec-Websocket-Protocol" header.
        """
        try:
            subprotocol, token = await get_subprotocol_and_token(sec_websocket_protocol)
        except WebsocketAuthError as exc:
            logger.info("Reject websocket client with invalid authentication", error=str(exc))
            await websocket_manager.disconnect(websocket, code=status.WS_1002_PROTOCOL_ERROR)
            return

        error = await websocket_manager.authorize(websocket, token)

        await websocket.accept(subprotocol=subprotocol)
        if error:
            await websocket_manager.disconnect(websocket, reason=error, code=status.WS_1008_POLICY_VIOLATION)
            return

        await websocket_manager.connect(websocket, WS_CHANNELS.EVENTS)



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Annotated

from authlib.integrations.starlette_client import OAuth
from fastapi import Depends
from fastapi.security.http import HTTPAuthorizationCredentials
from starlette.requests import Request
from starlette.websockets import WebSocket

from nwastdlib.url import URL
from oauth2_lib.fastapi import HTTPX_SSL_CONTEXT, HttpBearerExtractor, OIDCUserModel
from oauth2_lib.settings import oauth2lib_settings

oauth_client_credentials = OAuth()

oauth_client_credentials.register(
    "connext",
    server_metadata_url=URL(oauth2lib_settings.OIDC_CONF_URL),
    client_id=oauth2lib_settings.OAUTH2_RESOURCE_SERVER_ID,
    client_secret=oauth2lib_settings.OAUTH2_RESOURCE_SERVER_SECRET,
    request_token_params={"grant_type": "client_credentials"},
    client_kwargs={"verify": HTTPX_SSL_CONTEXT},
)

http_bearer_extractor = HttpBearerExtractor(auto_error=False)


async def authenticate(
    request: Request,
    http_auth_credentials: Annotated[HTTPAuthorizationCredentials | None, Depends(http_bearer_extractor)] = None,
) -> OIDCUserModel | None:
    token = http_auth_credentials.credentials if http_auth_credentials else None
    return await request.app.auth_manager.authentication.authenticate(request, token)


async def authorize(request: Request, user: Annotated[OIDCUserModel | None, Depends(authenticate)]) -> bool | None:
    return await request.app.auth_manager.authorization.authorize(request, user)


async def authenticate_websocket(websocket: WebSocket, token: str) -> OIDCUserModel | None:
    return await websocket.app.auth_manager.authentication.authenticate(websocket, token)


async def authorize_websocket(
    websocket: WebSocket, user: Annotated[OIDCUserModel | None, Depends(authenticate)]
) -> bool | None:
    return await websocket.app.auth_manager.authorization.authorize(websocket, user)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
from typing import Any

from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore
from apscheduler.schedulers.background import BackgroundScheduler
from more_itertools import partition
from pydantic import BaseModel

from orchestrator.db.filters import Filter
from orchestrator.db.filters.filters import CallableErrorHandler
from orchestrator.db.sorting import Sort
from orchestrator.db.sorting.sorting import SortOrder
from orchestrator.settings import app_settings
from orchestrator.utils.helpers import camel_to_snake, to_camel

jobstores = {"default": SQLAlchemyJobStore(url=str(app_settings.DATABASE_URI))}

scheduler = BackgroundScheduler(jobstores=jobstores)


def scheduler_dispose_db_connections() -> None:
    jobstores["default"].engine.dispose()


class ScheduledTask(BaseModel):
    id: str
    name: str | None = None
    next_run_time: datetime | None = None
    trigger: str


scheduled_task_keys = set(ScheduledTask.model_fields.keys())
scheduled_task_filter_keys = sorted(scheduled_task_keys | {to_camel(key) for key in scheduled_task_keys})
scheduled_task_sort_keys = scheduled_task_filter_keys


def scheduled_task_in_filter(job: ScheduledTask, filter_by: list[Filter]) -> bool:
    return any(f.value.lower() in getattr(job, camel_to_snake(f.field), "").lower() for f in filter_by)


def filter_scheduled_tasks(
    scheduled_tasks: list[ScheduledTask],
    handle_filter_error: CallableErrorHandler,
    filter_by: list[Filter] | None = None,
) -> list[ScheduledTask]:
    if not filter_by:
        return scheduled_tasks

    try:
        invalid_filters, valid_filters = partition(lambda x: x.field in scheduled_task_filter_keys, filter_by)

        if invalid_list := [item.field for item in invalid_filters]:
            handle_filter_error(
                "Invalid filter arguments", invalid_filters=invalid_list, valid_filter_keys=scheduled_task_filter_keys
            )

        valid_filter_list = list(valid_filters)
        return [task for task in scheduled_tasks if scheduled_task_in_filter(task, valid_filter_list)]
    except Exception as e:
        handle_filter_error(str(e))
        return []


def _invert(value: Any) -> Any:
    """Invert value for descending order."""
    if isinstance(value, (int, float)):
        return -value
    if isinstance(value, str):
        return tuple(-ord(c) for c in value)
    if isinstance(value, datetime):
        return -value.timestamp()
    return value


def sort_key(sort_field: str, sort_order: SortOrder) -> Any:
    def _sort_key(task: Any) -> Any:
        value = getattr(task, camel_to_snake(sort_field), None)
        if sort_field == "next_run_time" and value is None:
            return float("inf") if sort_order == SortOrder.ASC else float("-inf")
        return value if sort_order == SortOrder.ASC else _invert(value)

    return _sort_key


def sort_scheduled_tasks(
    scheduled_tasks: list[ScheduledTask], handle_sort_error: CallableErrorHandler, sort_by: list[Sort] | None = None
) -> list[ScheduledTask]:
    if not sort_by:
        return scheduled_tasks

    try:
        invalid_sorting, valid_sorting = partition(lambda x: x.field in scheduled_task_sort_keys, sort_by)
        if invalid_list := [item.field for item in invalid_sorting]:
            handle_sort_error(
                "Invalid sort arguments", invalid_sorting=invalid_list, valid_sort_keys=scheduled_task_sort_keys
            )

        valid_sort_list = list(valid_sorting)
        return sorted(
            scheduled_tasks, key=lambda task: tuple(sort_key(sort.field, sort.order)(task) for sort in valid_sort_list)
        )
    except Exception as e:
        handle_sort_error(str(e))
        return []


def default_error_handler(message: str, **context) -> None:  # type: ignore
    from orchestrator.graphql.utils.create_resolver_error_handler import _format_context

    raise ValueError(f"{message} {_format_context(context)}")


def get_scheduler_tasks(
    first: int = 10,
    after: int = 0,
    filter_by: list[Filter] | None = None,
    sort_by: list[Sort] | None = None,
    error_handler: CallableErrorHandler = default_error_handler,
) -> tuple[list[ScheduledTask], int]:
    scheduler.start(paused=True)
    scheduled_tasks = scheduler.get_jobs()
    scheduler.shutdown(wait=False)
    scheduler_dispose_db_connections()

    scheduled_tasks = filter_scheduled_tasks(scheduled_tasks, error_handler, filter_by)
    scheduled_tasks = sort_scheduled_tasks(scheduled_tasks, error_handler, sort_by)

    total = len(scheduled_tasks)
    paginated_tasks = scheduled_tasks[after : after + first + 1]

    return [
        ScheduledTask(
            id=task.id,
            name=task.name,
            next_run_time=task.next_run_time,
            trigger=str(task.trigger),
        )
        for task in paginated_tasks
    ], total



from orchestrator.db.filters.filters import (
    CallableErrorHandler,
    Filter,
    QueryType,
    create_memoized_field_list,
    generic_filter_from_clauses,
    generic_filters_validate,
)

__all__ = [
    "Filter",
    "CallableErrorHandler",
    "QueryType",
    "create_memoized_field_list",
    "generic_filter_from_clauses",
    "generic_filters_validate",
]



import re
from pathlib import Path
from subprocess import run
from typing import Callable


def remove_imports(text: str, module: str, symbol: str) -> tuple[str, bool]:
    """Find imports and remove them.

    Assumes code is formatted through Black to keep the regex somewhat readable.
    """
    text_orig = text

    # single import from module (may have a #comment) -> remove line
    rgx = r"(from %s import \b%s\b(\s*#[^\n]*)*\n)" % (re.escape(module), symbol)
    text = re.sub(rgx, "", text)

    # middle or last of multiple imports from module -> strip symbol
    rgx = r"(from %s import .+)(, \b%s\b)" % (re.escape(module), symbol)
    text = re.sub(rgx, r"\1", text)

    # first of multiple imports from same module -> strip symbol
    rgx = r"(from %s import )\b%s\b, " % (re.escape(module), symbol)
    text = re.sub(rgx, r"\1", text)

    # multiline import -> remove line with symbol
    rgx_verbose = r"""(?P<before>^from\s%s\simport\s*\([^\n]*\n(?:^[^\n]+,\n)*)
                      (^\s*\b%s\b,[^\n]*\n)
                      (?P<after>(?:^[^\n]+,\n)*\)[^\n]*$)"""
    text = re.sub(rgx_verbose % (re.escape(module), symbol), r"\g<before>\g<after>", text, flags=re.M | re.X)
    return text, text_orig != text


def insert_import(text: str, import_stmt: str) -> str:
    # Find the first import line and add our line above that
    # Rely on ruff & black for formatting
    return re.sub(r"(^(?:from .+|import .+)$)", f"{import_stmt}\n" + r"\1", text, count=1, flags=re.M)


def move_import(f: Path, symbol: str, old_module: str, new_module: str) -> bool:
    text = f.read_text()
    text, changed = remove_imports(text, old_module, symbol)
    if not changed:
        return False
    text = insert_import(text, f"from {new_module} import {symbol}")
    with f.open(mode="w"):
        f.write_text(text)
    return True


def find_and_remove_aliases(text: str, symbol: str) -> tuple[str, list[str]]:
    """In the given text find aliases of the given symbol and remove them.

    Return updated text and aliases removed.
    """
    rgx = r"(\b%s as (\w+))" % (symbol,)
    aliases = [aliasgroup for fullgroup, aliasgroup in re.findall(rgx, text)]
    newtext = re.sub(rgx, symbol, text)
    return newtext, aliases


def replace_words(text: str, words: list[str], replace: str) -> str:
    rgx = r"\b(%s)\b" % ("|".join(words),)
    return re.sub(rgx, replace, text)


def has_word(text: str, word: str) -> bool:
    return bool(re.search(r"\b%s\b" % (word,), text))


def run_tool(*args: str) -> bool:
    cmd = " ".join(args)
    try:
        r = run(args, capture_output=True)  # noqa: S603
        if r.returncode == 0:
            return True
        print(f"{cmd} failed:", r.stdout, r.stderr)
    except FileNotFoundError:
        print(f"{cmd} failed: could not find executable in the current venv")
    return False


def migrate(target_dir: Path, migrate_file: Callable[[Path], bool]) -> bool:
    abs_path = str(target_dir.resolve())

    def run_tools() -> bool:
        return run_tool("ruff", "check", "--fix", abs_path) and run_tool("black", "--quiet", abs_path)

    print(f"\n### Verifying files in {abs_path}... ", end="")
    if not run_tools():
        print("Failed to verify files, aborting migration. Please resolve errors.")
        return False
    print("Ok")

    files_migrated = files_checked = 0
    print(f"\n### Migrating files in {abs_path}")
    try:
        for f in target_dir.glob("**/*.py"):
            if migrate_file(f):
                files_migrated += 1
            files_checked += 1
    except KeyboardInterrupt:
        print("Interrupted...")

    print(f"\n### Migrated {files_migrated}/{files_checked} files in {abs_path}")

    print(f"\n### Formatting files in {abs_path}")
    return run_tools()



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Annotated, Any
from uuid import UUID

from pydantic import ConfigDict, Field

from orchestrator.config.assignee import Assignee
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.subscription import SubscriptionSchema
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus


class ProcessIdSchema(OrchestratorBaseModel):
    id: UUID


class ProcessBaseSchema(OrchestratorBaseModel):
    process_id: UUID
    workflow_id: UUID
    workflow_name: str
    is_task: bool
    created_by: str | None = None
    failed_reason: str | None = None
    started_at: datetime
    last_status: ProcessStatus
    last_step: str | None = None
    assignee: Assignee
    last_modified_at: datetime
    traceback: str | None = None
    model_config = ConfigDict(from_attributes=True)


class ProcessStepSchema(OrchestratorBaseModel):
    step_id: UUID | None = None
    name: str
    status: str
    created_by: str | None = None
    executed: datetime | None = Field(
        None, deprecated="Deprecated, use 'started' and 'completed' for step start and completion times"
    )
    started: datetime | None = None
    completed: datetime | None = None
    commit_hash: str | None = None
    state: dict[str, Any] | None = None
    state_delta: dict[str, Any] | None = None


class ProcessSchema(ProcessBaseSchema):
    product_id: UUID | None = None
    customer_id: str | None = None
    workflow_target: Target | None = None
    subscriptions: list[SubscriptionSchema]
    current_state: dict[str, Any] | None = None
    steps: list[ProcessStepSchema] | None = None
    form: dict[str, Any] | None = None


class ProcessResumeAllSchema(OrchestratorBaseModel):
    count: int


class ProcessStatusCounts(OrchestratorBaseModel):
    process_counts: dict[ProcessStatus, int]
    task_counts: dict[ProcessStatus, int]


Reporter = Annotated[str, Field(max_length=100)]



import structlog

from orchestrator.db.filters import Filter
from orchestrator.db.sorting import Sort
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.scheduled_task import ScheduledTaskGraphql
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.graphql.utils import create_resolver_error_handler, to_graphql_result_page
from orchestrator.graphql.utils.is_query_detailed import is_querying_page_data
from orchestrator.schedules.scheduler import get_scheduler_tasks, scheduled_task_filter_keys, scheduled_task_sort_keys

logger = structlog.get_logger(__name__)


async def resolve_scheduled_tasks(
    info: OrchestratorInfo,
    filter_by: list[GraphqlFilter] | None = None,
    sort_by: list[GraphqlSort] | None = None,
    first: int = 10,
    after: int = 0,
) -> Connection[ScheduledTaskGraphql]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    scheduled_tasks, total = get_scheduler_tasks(
        first=first, after=after, filter_by=pydantic_filter_by, sort_by=pydantic_sort_by, error_handler=_error_handler
    )

    graphql_scheduled_tasks = []
    if is_querying_page_data(info):
        graphql_scheduled_tasks = [ScheduledTaskGraphql.from_pydantic(p) for p in scheduled_tasks]

    return to_graphql_result_page(
        graphql_scheduled_tasks, first, after, total, scheduled_task_filter_keys, scheduled_task_sort_keys
    )



from collections.abc import Generator
from itertools import chain
from typing import Any

from more_itertools import flatten
from sqlalchemy import select
from sqlalchemy.sql.expression import Delete, Insert
from sqlalchemy.sql.selectable import ScalarSelect

from orchestrator.cli.domain_gen_helpers.helpers import (
    format_block_relation_to_dict,
    get_product_block_names,
    sql_compile,
)
from orchestrator.cli.domain_gen_helpers.types import BlockRelationDict, DomainModelChanges
from orchestrator.cli.helpers.input_helpers import get_user_input
from orchestrator.cli.helpers.print_helpers import COLOR, print_fmt, str_fmt
from orchestrator.db import db
from orchestrator.db.models import (
    ProductBlockRelationTable,
    ProductBlockTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
)
from orchestrator.domain.base import ProductBlockModel, get_depends_on_product_block_type_list


def get_product_block_id(block_name: str) -> ScalarSelect:
    return get_product_block_ids([block_name])


def get_product_block_ids(block_names: list[str] | set[str]) -> ScalarSelect:
    return select(ProductBlockTable.product_block_id).where(ProductBlockTable.name.in_(block_names)).scalar_subquery()


def get_subscription_instance(subscription_id: str, product_block_id: ScalarSelect) -> ScalarSelect:
    return (
        select(SubscriptionInstanceTable.subscription_instance_id)
        .where(
            SubscriptionInstanceTable.product_block_id.in_(product_block_id),
            SubscriptionInstanceTable.subscription_id == subscription_id,
        )
        .limit(1)
        .as_scalar()
    )


def map_create_product_blocks(product_blocks: dict[str, type[ProductBlockModel]]) -> dict[str, type[ProductBlockModel]]:
    """Map product blocks to create.

    Args:
        product_blocks: Dict of product blocks mapped by product block name.

    Returns: Dict of product blocks by product block name to create.
    """
    _existing_product_blocks = db.session.scalars(select(ProductBlockTable.name))
    existing_product_blocks = set(_existing_product_blocks)
    return {
        block_name: block for block_name, block in product_blocks.items() if block_name not in existing_product_blocks
    }


def map_delete_product_blocks(product_blocks: dict[str, type[ProductBlockModel]]) -> set[str]:
    """Map product blocks to delete.

    Args:
        product_blocks: Dict of product blocks mapped by product block name.

    Returns: List of product block names to delete.
    """
    existing_product_blocks = db.session.scalars(select(ProductBlockTable.name))
    return {name for name in existing_product_blocks if name not in product_blocks}


def map_product_block_additional_relations(
    changes: DomainModelChanges, models: dict[str, type[ProductBlockModel]], confirm_warnings: bool
) -> DomainModelChanges:
    """Map additional relations for created product blocks.

    Adds resource type and product block relations.

    Args:
        changes: DomainModelChanges class with all changes.
        models: All product block models.
        confirm_warnings: confirm warnings to continue, fully knowing that things can go wrong.

    Returns: Updated DomainModelChanges.
    """

    for block_name, block_class in changes.create_product_blocks.items():
        for field_name in block_class._non_product_block_fields_.keys():
            changes.create_resource_type_relations.setdefault(field_name, set()).add(block_name)

        product_blocks_in_model = block_class._get_depends_on_product_block_types()
        product_blocks_types_in_model = get_depends_on_product_block_type_list(product_blocks_in_model)
        for product_block_name in get_product_block_names(product_blocks_types_in_model):
            relation_list = changes.create_product_block_relations.get(product_block_name, [])
            new_relation = format_block_relation_to_dict(block_name, product_block_name, models, confirm_warnings)

            if new_relation not in relation_list:
                changes.create_product_block_relations[product_block_name] = relation_list + [new_relation]

    return changes


def generate_create_product_blocks_sql(
    create_product_blocks: dict[str, Any], inputs: dict[str, dict[str, str]]
) -> list[str]:
    """Generate SQL to create product blocks.

    Args:
        create_product_blocks: List of product block names.
        inputs: Optional Dict to prefill 'description' and 'tag' per product block.
            - key: product block name.
            - value: Dict with 'description' and 'tag'.
                - key: product block property.
                - value: value for the property.

    Returns: List of SQL to create product blocks.
    """
    print_fmt("\nCreate product blocks", flags=[COLOR.BOLD, COLOR.UNDERLINE])

    def create_product_block(name: str) -> str:
        print(f"Product block: {str_fmt(name, flags=[COLOR.BOLD])}")  # noqa: T001, T201
        prefilled_values = inputs.get(name, {})
        description = prefilled_values.get("description") or get_user_input("Supply the product block description: ")
        tag = prefilled_values.get("tag") or get_user_input("Supply the product block tag: ")
        return sql_compile(
            Insert(ProductBlockTable).values(
                {
                    "name": name,
                    "description": description,
                    "tag": tag,
                    "status": "active",
                }
            )
        )

    return [create_product_block(name) for name in create_product_blocks]


def generate_delete_product_blocks_sql(delete_product_blocks: set[str]) -> list[str]:
    """Generate SQL to delete product blocks.

    Args:
        delete_product_blocks: List of product block names.

    Returns: List of SQL to delete product blocks.
    """

    if not delete_product_blocks:
        return []
    return [
        sql_compile(
            Delete(SubscriptionInstanceTable).where(
                SubscriptionInstanceTable.product_block_id.in_(get_product_block_ids(delete_product_blocks))
            )
        ),
        sql_compile(Delete(ProductBlockTable).where(ProductBlockTable.name.in_(delete_product_blocks))),
    ]


def generate_create_product_block_relations_sql(
    create_block_relations: dict[str, list[BlockRelationDict]],
) -> list[str]:
    """Generate SQL to create product block to product block relations.

    Args:
        create_block_relations: Dict with product blocks by product block
            - key: product block name.
            - list: List of product blocks to relate with by prop names.

    Returns: List of SQL to create relation between product blocks.
    """

    def create_block_relation(depends_block_name: str, block_relations: list[BlockRelationDict]) -> str:
        depends_block_id_sql = get_product_block_id(depends_block_name)

        def create_block_relation_dict(block_relation: BlockRelationDict) -> dict[str, ScalarSelect]:
            block_id_sql = get_product_block_id(block_relation["name"])
            return {"in_use_by_id": block_id_sql, "depends_on_id": depends_block_id_sql}

        product_product_block_relation_dicts = [create_block_relation_dict(block) for block in block_relations]
        return sql_compile(Insert(ProductBlockRelationTable).values(product_product_block_relation_dicts))

    return [create_block_relation(*item) for item in create_block_relations.items()]


def generate_create_product_block_instance_relations_sql(
    product_block_relations: dict[str, list[BlockRelationDict]],
) -> list[str]:
    """Generate SQL to create resource type instance values for existing instances.

    Args:
        product_block_relations: Dict with product blocks by resource type
            - key: product block name.
            - value: Set of product block names to relate to.

    Returns: .
    """

    def create_subscription_instance_relations(
        depends_block_name: str, block_relations: list[BlockRelationDict]
    ) -> Generator[str, None, None]:
        depends_block_id_sql = get_product_block_id(depends_block_name)

        def map_subscription_instances(
            block_relation: BlockRelationDict,
        ) -> dict[str, list[dict[str, str | ScalarSelect]]]:
            in_use_by_id_sql = get_product_block_id(block_relation["name"])
            stmt = select(
                SubscriptionInstanceTable.subscription_instance_id, SubscriptionInstanceTable.subscription_id
            ).where(SubscriptionInstanceTable.product_block_id.in_(in_use_by_id_sql))

            subscription_instances = list(db.session.execute(stmt))
            if not subscription_instances:
                subscription_instances = []

            instance_list = [
                {"subscription_id": subscription_id, "product_block_id": depends_block_id_sql}
                for instance_id, subscription_id, in subscription_instances
            ]
            instance_relation_list = [
                {
                    "in_use_by_id": instance.subscription_instance_id,
                    "depends_on_id": get_subscription_instance(instance.subscription_id, depends_block_id_sql),
                    "order_id": 0,
                    "domain_model_attr": block_relation["attribute_name"],
                }
                for instance in subscription_instances
            ]

            return {"instance_list": instance_list, "instance_relation_list": instance_relation_list}

        create_instance_list = [map_subscription_instances(block_relation) for block_relation in block_relations]

        subscription_instance_dicts = list(flatten(item["instance_list"] for item in create_instance_list))
        subscription_relation_dicts = list(flatten(item["instance_relation_list"] for item in create_instance_list))

        if subscription_instance_dicts:
            yield sql_compile(Insert(SubscriptionInstanceTable).values(subscription_instance_dicts))

        if subscription_relation_dicts:
            yield sql_compile(Insert(SubscriptionInstanceRelationTable).values(subscription_relation_dicts))

    return list(
        chain.from_iterable(create_subscription_instance_relations(*item) for item in product_block_relations.items())
    )


def generate_delete_product_block_relations_sql(delete_block_relations: dict[str, set[str]]) -> list[str]:
    """Generate SQL to delete product block to product blocks relations.

    Args:
        delete_block_relations: Dict with product blocks by product block
            - key: Product block name.
            - value: Set of product block names to relate with.

    Returns: List of SQL to delete relations between product blocks.
    """

    def delete_block_relation(delete_block_name: str, block_names: set[str]) -> str:
        in_use_by_ids_sql = get_product_block_ids(block_names)
        delete_block_id_sql = get_product_block_id(delete_block_name)

        return sql_compile(
            Delete(ProductBlockRelationTable).where(
                ProductBlockRelationTable.in_use_by_id.in_(in_use_by_ids_sql),
                ProductBlockRelationTable.depends_on_id.in_(delete_block_id_sql),
            )
        )

    return [delete_block_relation(*item) for item in delete_block_relations.items()]



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from orchestrator.graphql.schemas.product import ProductModelGraphql
from orchestrator.graphql.schemas.strawberry_pydantic_patch import (
    convert_pydantic_model_to_strawberry_class__patched,  # noqa: F401
)
from orchestrator.graphql.types import StrawberryModelType

DEFAULT_GRAPHQL_MODELS: StrawberryModelType = {
    "ProductModelGraphql": ProductModelGraphql,
}
